import asyncio
import collections


class Pool:
    def __init__(self, factory, minsize=1, maxsize=10):
        self.factory = factory
        self.minsize = minsize
        self.maxsize = maxsize
        self._pool = collections.deque(maxlen=maxsize)
        self._cond = asyncio.Condition(asyncio.Lock())
        self._using = set()
        self._acquiring = 0
        self._close_state = asyncio.Event()
        self._fill_free(override_min=False)

    def __repr__(self):
        return '<{}: size={}>'.format(self.__class__.__name__, self.size)

    @property
    def size(self):
        return self.freesize + len(self._using) + self._acquiring

    @property
    def freesize(self):
        return len(self._pool)

    async def clear(self):
        async with self._cond:
            while self._pool:
                client = self._pool.popleft()
                await client.close()

    async def _do_close(self):
        async with self._cond:
            assert self._acquiring == 0, self._acquiring
            while self._pool:
                client = self._pool.popleft()
                await client.close()
            for client in self._using:
                await client.close()
            self._using.clear()

    async def close(self):
        if not self._close_state.is_set():
            self._close_state.set()
            await self._do_close()

    @property
    def closed(self):
        return self._close_state.is_set()

    async def acquire(self):
        async with self._cond:
            while True:
                self._fill_free(override_min=True)
                if self.freesize:
                    client = self._pool.popleft()
                    assert client not in self._using, (client, self._using)
                    self._using.add(client)
                    return client
                else:
                    await self._cond.wait()

    async def release(self, client):
        assert client in self._using
        self._using.remove(client)
        self._pool.append(client)
        await self._wakeup()

    async def _wakeup(self):
        async with self._cond:
            self._cond.notify()

    def _fill_free(self, *, override_min):
        while self.size < self.minsize:
            self._acquiring += 1
            try:
                client = self.factory()
                self._pool.append(client)
            finally:
                self._acquiring -= 1
        if self.freesize:
            return
        if override_min:
            while not self._pool and self.size < self.maxsize:
                self._acquiring += 1
                try:
                    client = self.factory()
                    self._pool.append(client)
                finally:
                    self._acquiring -= 1

    async def auto_release(self, client, coro):
        try:
            return await coro
        finally:
            await self.release(client)

    def get(self):
        return AsyncClientContextManager(self)


class AsyncClientContextManager:
    __slots__ = ('_pool', '_client')

    def __init__(self, pool):
        self._pool = pool
        self._client = None

    async def __aenter__(self):
        self._client = await self._pool.acquire()
        return self._client

    async def __aexit__(self, exc_type, exc_value, tb):
        try:
            await self._pool.release(self._client)
        finally:
            self._pool = None
            self._client = None



import time
import json
import logging
import aiobotocore
from .worker import BatchWorker
logger = logging.getLogger(__name__)


class Firehose:
    service_name = 'firehose'

    def __init__(self, stream_name, region_name,
                 encode_func=None, delimiter=b'\n',
                 *, worker_params=None, client_params=None):
        self.session = aiobotocore.get_session()
        self.stream_name = stream_name
        self.encode_func = encode_func or (lambda o: json.dumps(o).encode())
        self.delimiter = delimiter
        client_params = client_params or {}
        client_params['region_name'] = region_name
        self.client = self.session.create_client(
            self.service_name, **client_params)
        worker_params = worker_params or {}
        self.worker = BatchWorker(self.write_batch, **worker_params)

    def __repr__(self):
        return '<{}: stream={}, worker={!r}>'.format(
                self.__class__.__name__, self.stream_name, self.worker)

    async def put(self, obj):
        return await self.worker.put(obj)

    async def stop(self):
        timestamp = time.time()
        await self.worker.stop()
        await self.client.close()
        cost = time.time() - timestamp
        logger.info('{0!r} stopped in {1:.1f} seconds'.format(self, cost))

    def _encode(self, obj_list):
        encoded = [self.encode_func(obj) for obj in obj_list]
        encoded.append(b'')
        return self.delimiter.join(encoded)

    async def write_batch(self, obj_list):
        try:
            data = self._encode(obj_list)
            await self.client.put_record(
                DeliveryStreamName=self.stream_name,
                Record={'Data': data}
            )
        except Exception as e:
            logger.exception(e)
            return obj_list



import time
import base64
import pickle
import logging
import asyncio
import functools
import aiobotocore
from .worker import BatchWorker
loop = asyncio.get_event_loop()
logger = logging.getLogger(__name__)


class SQSWriter:
    service_name = 'sqs'
    MAX_RETRY = 10

    def __init__(self, queue_name, region_name, encode_func=None,
                 *, worker_params=None, client_params=None):
        self.session = aiobotocore.get_session()
        self.queue_name = queue_name
        self.is_fifo = queue_name.endswith('.fifo')
        self.encode_func = encode_func or (
            lambda o: base64.b64encode(pickle.dumps(o, 2)).decode('ascii')
        )
        client_params = client_params or {}
        client_params['region_name'] = region_name
        self.client = self.session.create_client(
            self.service_name, **client_params)
        self.QueueUrl = None
        self.lock = asyncio.Lock()
        self.loop = asyncio.get_event_loop()
        worker_params = worker_params or {}
        self.worker = BatchWorker(self.write_batch, **worker_params)

    def __repr__(self):
        return '<{}: queue={}, worker={!r}>'.format(
                self.__class__.__name__, self.queue_name, self.worker)

    async def _get_queue_url(self):
        if self.QueueUrl is None:
            async with self.lock:
                resp = await self.client.get_queue_url(
                        QueueName=self.queue_name
                    )
                self.QueueUrl = resp['QueueUrl']
        return self.QueueUrl

    async def stop(self):
        timestamp = time.time()
        await self.worker.stop()
        await self.client.close()
        cost = time.time() - timestamp
        logger.info('{0!r} stopped in {1:.1f} seconds'.format(self, cost))

    def _encode(self, obj):
        return self.encode_func(obj)

    async def write_one(self, record, delay_seconds=0,
                        deduplication_id=None, group_id=None,
                        queued=False, _seq=0, **attributes):
        if _seq > 0:
            await asyncio.sleep(0.1 * (2 ** _seq))
        if _seq > self.MAX_RETRY - 1:
            raise Exception('write_one error: SQS send_message failed')
        message = {
            'record': record,
            'delay_seconds': delay_seconds,
            'attributes': attributes,
            'params': {},
        }
        if self.is_fifo:
            if deduplication_id:
                message['params']['MessageDeduplicationId'] = deduplication_id
            if group_id:
                message['params']['MessageGroupId'] = group_id
        if queued:
            await self.worker.put(message)
            return
        try:
            return await self.client.send_message(
                QueueUrl=(await self._get_queue_url()),
                MessageBody=self._encode(record),
                DelaySeconds=delay_seconds,
                MessageAttributes=attributes,
                **message['params'],
            )
        except Exception as e:
            logger.exception(e)
            return await self.write_one(
                    record, delay_seconds=delay_seconds,
                    deduplication_id=deduplication_id, group_id=group_id,
                    queued=False, _seq=_seq+1, **attributes)

    async def write_batch(self, obj_list, _seq=0):
        if _seq > 0:
            await asyncio.sleep(0.1 * (2 ** _seq))
        if _seq > self.MAX_RETRY - 1:
            raise Exception('write_batch error: SQS send_message_batch failed',
                            obj_list)
        Entries = [
            {
                'Id': str(i),
                'MessageBody': self._encode(message['record']),
                'DelaySeconds': message['delay_seconds'],
                'MessageAttributes': message['attributes'],
                **message['params'],
            } for i, message in enumerate(obj_list)
        ]
        try:
            resp = await self.client.send_message_batch(
                QueueUrl=(await self._get_queue_url()),
                Entries=Entries
            )
        except Exception as e:
            logger.exception(e)
            return obj_list
        if 'Failed' in resp:
            logger.error('Send failed: {}, {}'.format(
                obj_list, resp['Failed']))
            failed_obj_list = [
                obj_list[int(d['Id'])] for d in resp['Failed']
                if not d['SenderFault']
            ]
            return await self.write_batch(
                failed_obj_list, _seq=_seq + 1
            )

    def async_rpc(self, func=None, *, delay_seconds=0,
                  deduplication_id=None, group_id=None,
                  queued=True, **attributes):
        if func is None:
            return functools.partial(self.async_rpc,
                                     delay_seconds=delay_seconds,
                                     deduplication_id=deduplication_id,
                                     group_id=group_id,
                                     queued=queued, **attributes)
        if type(func) == str:
            async def async_func(*args, **kw):
                record = {
                    'fpath': func,
                    'args': args,
                    'kw': kw,
                }
                await self.write_one(record, delay_seconds=delay_seconds,
                                     deduplication_id=deduplication_id,
                                     group_id=group_id,
                                     queued=queued, **attributes)
        else:
            @functools.wraps(func)
            async def async_func(*args, **kw):
                fpath = func.__module__ + '.' + func.__name__
                record = {
                    'fpath': fpath,
                    'args': args,
                    'kw': kw,
                }
                await self.write_one(record, delay_seconds=delay_seconds,
                                     deduplication_id=deduplication_id,
                                     group_id=group_id,
                                     queued=queued, **attributes)
        return async_func


class StringAttribute(dict):
    def __init__(self, value):
        self['DataType'] = 'String'
        self['StringValue'] = str(value)


class BinaryAttribute(dict):
    def __init__(self, value):
        self['DataType'] = 'Binary'
        self['BinaryValue'] = bytes(value)


class NumberAttribute(dict):
    def __init__(self, value):
        self['DataType'] = 'Number'
        self['StringValue'] = str(value)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import os.path
import re
VERSION_RE = re.compile(r'''__version__ = ['"]([-a-z0-9.]+)['"]''')
BASE_PATH = os.path.dirname(__file__)


with open(os.path.join(BASE_PATH, 'cowfish', '__init__.py')) as f:
    try:
        version = VERSION_RE.search(f.read()).group(1)
    except IndexError:
        raise RuntimeError('Unable to determine version.')


setup(
    name='cowfish',
    description='A useful asynchronous library bases on aiobotocore',
    license='MIT',
    version=version,
    author='Yingbo Gu',
    author_email='tensiongyb@gmail.com',
    maintainer='Yingbo Gu',
    maintainer_email='tensiongyb@gmail.com',
    url='https://github.com/guyingbo/cowfish',
    packages=['cowfish'],
    python_requires='>=3.5',
    install_requires=[
        'aiobotocore>=0.4.5',
    ],
    entry_points={
        'console_scripts': [
            'sqsprocesser = cowfish.sqsprocesser:main',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
)
